/*****************************************************************************
 *
 * Multitouch research project
 *
 * Copyright (C) 2012 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/

#include <stdio.h>
#include "Device.h"

Device::Device() : id(-1) {
}

Device::Device(int id_) : id(id_) {
}

Device::~Device() {
}

/*
 * Check that all old (non-ending) touches are represented in the new frame.
 */

bool Device::check_validity_forward(const Frame &new_frame, const std::vector<Touch> &old_touches) {
    bool result = true;
    for(size_t i=0; i<old_touches.size(); i++) {
        const Touch &old_touch = old_touches[i];
        enum TouchState old_state = old_touch.get_state();
        const Touch &new_touch = new_frame.get_touch_id(old_touch.get_id());
        enum TouchState new_state = new_touch.get_state();

        switch(old_state) {
        case TouchBegin:
        case TouchUpdate:
            if(!(new_state == TouchUpdate || new_state == TouchEnd)) {
                if(new_touch.is_invalid())
                    printf("Lost touch %d in frame %d.\n", old_touch.get_id(), new_frame.get_id());
                else
                    printf("Bad state for existing touch. New touch state: %s.\n", state_to_string(new_state));
                result = false;
            }
            break;
        case TouchEnd:
            if (!(new_touch.is_invalid() || new_state == TouchBegin)) {
                printf("Invalid continuation of an ended touch. New state %s.\n", state_to_string(new_state));
                result = false;
            }
            break;
        case TouchUnknown :
            break; // Don't know what to do here, really.
        }

    }
    return result;
}

/* Check that all (non-starting) touches exist in the previous frame. */

bool Device::check_validity_backward(const Frame &old_frame, const std::vector<Touch> &new_touches) {
    bool result = true;
    for(size_t i=0; i<new_touches.size(); i++) {
        const Touch &new_touch = new_touches[i];
        enum TouchState new_state = new_touch.get_state();
        const Touch &old_touch = old_frame.get_touch_id(new_touch.get_id());
        enum TouchState old_state = old_touch.get_state();

        switch(new_state) {
        case TouchEnd:
        case TouchUpdate:
            if(!(old_state == TouchUpdate || old_state == TouchBegin)) {
                if(old_touch.is_invalid())
                    printf("Touch %d (in %s) without previous state in frame %d.\n", new_touch.get_id(), state_to_string(new_state), old_frame.get_id());
                else
                    printf("Bad state for existing touch. New touch state: %s.\n", state_to_string(new_state));
                result = false;
            }
            break;
        case TouchBegin:
            if (!(old_touch.is_invalid() || old_state == TouchEnd)) {
                printf("TouchBegin for an existing touch. Old state %s.\n", state_to_string(old_state));
                result = false;
            }
            break;
        case TouchUnknown :
            break; // Don't know what to do here, really.
        }

    }
    return result;
}

bool Device::is_new_frame_valid(const Frame &new_frame) {
    if(frames.size() == 0)
        return true;
    const Frame &old_frame = frames.back();
    const std::vector<Touch> new_touches = new_frame.get_touches();
    const std::vector<Touch> old_touches = old_frame.get_touches();
    bool result = check_validity_forward(new_frame, old_touches) && check_validity_backward(old_frame, new_touches);

    return result;
}


void Device::add_frame(const Frame &f) {
    if(!is_new_frame_valid(f)) {
        printf("Touch states in new frame are inconsistent with previous frame.\n");
    }
    frames.push_back(f);
}

void Device::pop_oldest_frame() {
    if(frames.size() > 0)
        frames.erase(frames.begin());
}
