/*****************************************************************************
 *
 * Multitouch research project
 *
 * Copyright (C) 2012 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 ****************************************************************************/


#include <map>
#include <math.h>
#include <unistd.h>
#include <time.h>
#include "Device.h"
#include <CoreFoundation/CoreFoundation.h>

/*
 * These definitions are based on reverse-engineering work presented here:
 *
 * http://steike.com/code/multitouch/
 *
 */

typedef struct { float x,y; } mtPoint;
typedef struct { mtPoint pos,vel; } mtReadout;

typedef struct {
    int frame;
    double timestamp;
    int identifier, state, unknown1, unknown2;
    mtReadout normalized; // Coordinates between 0 and 1.
    float size;
    int unknown3;
    float angle, majorAxis, minorAxis; // ellipsoid
    mtReadout unknown4;
    int unknown5[2];
    float unknown6;
} Finger;

typedef void *MTDeviceRef;
typedef int (*MTContactCallbackFunction)(int,Finger*,int,double,int);

std::map<int, Device*> devices;

TouchState osx_state_to_internal_state(int state) {
    switch(state) {
    case 1 : return TouchBegin;
    case 7 : return TouchEnd;
    default : return TouchUpdate;
    }
}

extern "C" {
MTDeviceRef MTDeviceCreateDefault();
void MTRegisterContactFrameCallback(MTDeviceRef, MTContactCallbackFunction);
void MTDeviceStart(MTDeviceRef, int);

int callback(int device_num, Finger *data, int nFingers, double timestamp, int frame) {
    Device *device;
    if(devices.find(device_num) == devices.end()) {
        device = new Device(device_num);
        devices[device_num] = device;
        printf("Created a new device with id %d.\n", device_num);
    } else {
        device = devices[device_num];
    }
    std::vector<Touch> current_touches;

    int frame_num;
    for (int i=0; i<nFingers; i++) {
        Finger *f = &data[i];
        Touch t(f->identifier, f->normalized.pos.x, f->normalized.pos.y, timestamp,
                f->angle, f->minorAxis, f->majorAxis, osx_state_to_internal_state(f->state));
        frame_num = f->frame; // It is the same on all touches.
        current_touches.push_back(t);
    }
    Frame current_frame(frame_num, current_touches);
    device->add_frame(current_frame);
    while(device->num_frames() > 100)
        device->pop_oldest_frame();
    printf("Now we iz have %d frames. The last one had %d touches.\n",
           (int)device->num_frames(), (int) device->get_frames().back().get_touches().size());
    return 0;
}

} // Extern "C"

int main() {
    MTDeviceRef dev = MTDeviceCreateDefault();
    MTRegisterContactFrameCallback(dev, callback);
    MTDeviceStart(dev, 0);
    printf("Ctrl-C to abort\n");
    sleep(-1);
    for(std::map<int, Device*>::iterator it = devices.begin(); it != devices.end(); ++it)
        delete it->second;
    devices.clear();
    return 0;
}

